from collections import OrderedDict
from andes.utils.func import list_flatten


# all_routines: file name: class name
all_routines = OrderedDict([('pflow', ['PFlow']),
                            ('tds', ['TDS']),
                            ('eig', ['EIG']),
                            ])

class_names = list_flatten(list(all_routines.values()))
routine_cli = OrderedDict([(item.lower(), item) for item in class_names])



"""
JSON reader and writer for ANDES.

"""
import json
import logging
from collections import OrderedDict
from andes.utils.paths import confirm_overwrite

logger = logging.getLogger(__name__)


def testlines(fid):
    return True


def write(system, outfile, skip_empty=True, overwrite=None, **kwargs):
    """
    Write loaded ANDES system data into a JSON file

    Parameters
    ----------
    system : System
        A loaded system with parameters
    outfile : str
        Path to the output file
    skip_empty : bool
        Skip output of empty models (n = 0)
    overwrite : bool
        None to prompt for overwrite selection; True to overwrite; False to not overwrite

    Returns
    -------
    bool
        True if file written; False otherwise
    """
    if not confirm_overwrite(outfile, overwrite):
        return False

    with open(outfile, 'w') as writer:
        writer.write(_dump_system(system, writer, skip_empty))
        logger.info(f'JSON file written to "{outfile}"')

    return True


def _dump_system(system, writer, skip_empty, orient='records'):
    """
    Write system to JSON output.
    """
    out = OrderedDict()
    for name, instance in system.models.items():
        if skip_empty and instance.n == 0:
            continue
        out[name] = instance.cache.df_in.to_dict(orient=orient)

    return json.dumps(out, indent=2)


def read(system, infile):
    """
    Read JSON file with ANDES model data into an empty system.

    Parameters
    ----------
    system : System
        Empty System instance
    infile : str
        Path to the input file

    Returns
    -------
    System
        System instance after succeeded
    """
    with open(infile, 'r') as f:
        json_in = json.load(f)

    for name, dct in json_in.items():
        for row in dct:
            system.add(name, row)

    return system



from typing import Optional, Union, List
from andes.core.param import BaseParam
from andes.core.common import DummyValue
from andes.core.discrete import Discrete
from andes.core.service import BaseService
from andes.models.group import GroupBase
from andes.shared import np, ndarray


class BaseVar(object):
    """
    Base variable class.

    Derived classes `State` and `Algeb` should be used to build
    model variables.

    Parameters
    ----------
    name : str, optional
        Variable name
    info : str, optional
        Descriptive information
    unit : str, optional
        Unit
    tex_name : str
        LaTeX-formatted variable name. If is None, use `name`
        instead.
    discrete : Discrete
        Associated discrete component. Will call `check_var` on
        the discrete component.

    Attributes
    ----------
    a : array-like
        variable address
    v : array-like
        local-storage of the variable value
    e : array-like
        local-storage of the corresponding equation value
    e_str : str
        the string/symbolic representation of the equation
    """
    def __init__(self,
                 name: Optional[str] = None,
                 tex_name: Optional[str] = None,
                 info: Optional[str] = None,
                 unit: Optional[str] = None,
                 v_str: Optional[Union[str, float]] = None,
                 v_iter: Optional[str] = None,
                 e_str: Optional[str] = None,
                 discrete: Optional[Discrete] = None,
                 v_setter: Optional[bool] = False,
                 e_setter: Optional[bool] = False,
                 addressable: Optional[bool] = True,
                 export: Optional[bool] = True,
                 diag_eps: Optional[float] = 0.0,
                 ):

        self.name = name
        self.info = info
        self.unit = unit

        self.tex_name = tex_name if tex_name else name
        self.owner = None  # instance of the owner Model
        self.id = None     # variable internal index inside a model (assigned in run time) FIXME: not in use

        self.v_str = v_str  # equation string (v = v_str) for variable initialization
        self.v_iter = v_iter  # the implicit equation (0 = v_iter) for iterative initialization
        self.e_str = e_str  # string for symbolic equation

        self.discrete = discrete
        self.v_setter = v_setter  # True if this variable sets the variable value
        self.e_setter = e_setter  # True if this var sets the equation value
        self.addressable = addressable  # True if this var needs to be assigned an address FIXME: not in use
        self.export = export            # True if this var's value needs to exported
        self.diag_eps = diag_eps        # small diagonal value (1e-6) to be added to `dae.gy`

        # attributes assigned by `set_address`
        self.n = 0
        self.a: ndarray = np.array([], dtype=int)       # address array
        self.v: ndarray = np.array([], dtype=np.float)  # variable value array
        self.e: ndarray = np.array([], dtype=np.float)  # equation value array

        self.av: ndarray = np.array([], dtype=int)      # FIXME: future var. address array
        self.ae: ndarray = np.array([], dtype=int)      # FIXME: future equation address array

        # internal flags
        # NOTE:
        # contiguous is only True for internal variables of models with flag ``collate`` equal to ``False``.
        self._contiguous = False  # True if if address is contiguous to allow slicing into arrays without copy.

        self.e_inplace = False    # True if `self.e` is in-place access to `System.dae.__dict__[self.e_code]`
        self.v_inplace = False    # True if `self.v` is in-place access to `System.dae.__dict__[self.v_code]`
        self.allow_none = False   # True to allow None in address (NOT IN USE)

    def reset(self):
        """
        Reset the internal numpy arrays and flags.
        """
        self.n = 0
        self.a[:] = 0
        self.v[:] = 0
        self.e[:] = 0
        self.av[:] = 0
        self.ae[:] = 0

        self._contiguous = False
        self.e_inplace = False
        self.v_inplace = False

    def __repr__(self):
        span = []
        if 1 <= self.n <= 20:
            span = self.a.tolist()
            span = ', '.join([str(i) for i in span])
        elif self.n > 20:
            if not isinstance(self, ExtVar):
                span.append(self.a[0])
                span.append(self.a[-1])
                span.append(self.a[1] - self.a[0])
                span = ':'.join([str(i) for i in span])

        if span:
            span = ' [' + span + ']'

        return f'{self.__class__.__name__}, {self.owner.__class__.__name__}.{self.name}{span}'

    def set_address(self, addr: ndarray, contiguous=False):
        """
        Set the address of internal variables.

        Parameters
        ----------
        addr : ndarray
            The assigned address for this variable
        contiguous : bool, optional
            If the addresses are contiguous
        """
        self.a = addr
        self.n = len(self.a)

        # NOT IN USE
        self.ae = np.array(self.a)
        self.av = np.array(self.a)
        # -----------

        self._contiguous = contiguous

        if self._contiguous:
            if self.e_setter is False:
                self.e_inplace = True

            if self.v_setter is False:
                self.v_inplace = True

    def set_arrays(self, dae):
        """
        Set the equation and values arrays.

        It slicing into DAE (when contiguous) or allocating new memory (when not contiguous).

        Parameters
        ----------
        dae : DAE
            Reference to System.dae
        """
        slice_idx = slice(self.a[0], self.a[-1] + 1)
        if self.v_inplace:
            self.v = dae.__dict__[self.v_code][slice_idx]
        else:
            self.v = np.zeros(self.n)

        if self.e_inplace:
            self.e = dae.__dict__[self.e_code][slice_idx]
        else:
            self.e = np.zeros(self.n)

    def get_names(self):
        return [self.name]

    @property
    def class_name(self):
        return self.__class__.__name__


class Algeb(BaseVar):
    """
    Algebraic variable class, an alias of the `BaseVar`.

    Attributes
    ----------
    e_code : str
        Equation code string, equals string literal ``g``
    v_code : str
        Variable code string, equals string literal ``y``
    """
    e_code = 'g'
    v_code = 'y'


class State(BaseVar):
    """
    Differential variable class, an alias of the `BaseVar`.

    Parameters
    ----------
    t_const : BaseParam, DummyValue
        Left-hand time constant for the differential equation.
        Time constants will not be evaluated as part of the differential equation.
        They will be collected to array `dae.Tf` to multiply to the right-hand side `dae.f`.

    Attributes
    ----------
    e_code : str
        Equation code string, equals string literal ``f``
    v_code : str
        Variable code string, equals string literal ``x``
    """
    e_code = 'f'
    v_code = 'x'

    def __init__(self,
                 name: Optional[str] = None,
                 tex_name: Optional[str] = None,
                 info: Optional[str] = None,
                 unit: Optional[str] = None,
                 v_str: Optional[Union[str, float]] = None,
                 v_iter: Optional[str] = None,
                 e_str: Optional[str] = None,
                 discrete: Optional[Discrete] = None,
                 t_const: Optional[Union[BaseParam, DummyValue]] = None,
                 v_setter: Optional[bool] = False,
                 e_setter: Optional[bool] = False,
                 addressable: Optional[bool] = True,
                 export: Optional[bool] = True,
                 diag_eps: Optional[float] = 0.0,
                 ):
        BaseVar.__init__(self, name=name,
                         tex_name=tex_name,
                         info=info,
                         unit=unit,
                         v_str=v_str,
                         v_iter=v_iter,
                         e_str=e_str,
                         discrete=discrete,
                         v_setter=v_setter,
                         e_setter=e_setter,
                         addressable=addressable,
                         export=export,
                         diag_eps=diag_eps,
                         )
        self.t_const = t_const


class ExtVar(BaseVar):
    """
    Externally defined algebraic variable

    This class is used to retrieve the addresses of externally-
    defined variable. The `e` value of the `ExtVar` will be added
    to the corresponding address in the DAE equation.

    Parameters
    ----------
    model : str
        Name of the source model
    src : str
        Source variable name
    indexer : BaseParam, BaseService
        A parameter of the hosting model, used as indices into
        the source model and variable. If is None, the source
        variable address will be fully copied.
    allow_none : bool
        True to allow None in indexer

    Attributes
    ----------
    parent_model : Model
        The parent model providing the original parameter.
    uid : array-like
        An array containing the absolute indices into the
        parent_instance values.
    e_code : str
        Equation code string; copied from the parent instance.
    v_code : str
        Variable code string; copied from the parent instance.
    """
    def __init__(self,
                 model: str,
                 src: str,
                 indexer: Optional[Union[List, ndarray, BaseParam, BaseService]] = None,
                 allow_none: Optional[bool] = False,
                 name: Optional[str] = None,
                 tex_name: Optional[str] = None,
                 info: Optional[str] = None,
                 unit: Optional[str] = None,
                 v_str: Optional[Union[str, float]] = None,
                 v_iter: Optional[str] = None,
                 e_str: Optional[str] = None,
                 v_setter: Optional[bool] = False,
                 e_setter: Optional[bool] = False,
                 addressable: Optional[bool] = True,
                 export: Optional[bool] = True,
                 diag_eps: Optional[float] = 0.0,
                 ):
        super().__init__(name=name,
                         tex_name=tex_name,
                         info=info,
                         unit=unit,
                         v_str=v_str,
                         v_iter=v_iter,
                         e_str=e_str,
                         v_setter=v_setter,
                         e_setter=e_setter,
                         addressable=addressable,
                         export=export,
                         diag_eps=diag_eps,
                         )
        self.model = model
        self.src = src
        self.indexer = indexer
        self.allow_none = allow_none
        self.parent = None
        self._idx = None
        self._n = []
        self._n_count = 0

    @property
    def _v(self):
        out = []
        idx = 0
        for n in self._n:
            out.append(self.v[idx:idx+n])
            idx += n
        return out

    @property
    def _a(self):
        out = []
        idx = 0
        for n in self._n:
            out.append(self.a[idx:idx+n])
            idx += n
        return out

    def set_address(self, addr, contiguous=False):
        """
        Empty function.
        """
        pass

    def set_arrays(self, dae):
        """Empty function."""
        pass

    def link_external(self, ext_model):
        """
        Update variable addresses provided by external models

        This method sets attributes including `parent_model`,
        `parent_instance`, `uid`, `a`, `n`, `e_code` and
        `v_code`. It initializes the `e` and `v` to zero.

        Returns
        -------
        None

        Parameters
        ----------
        ext_model : Model
            Instance of the parent model

        Warnings
        --------
        `link_external` does not check if the ExtVar type is the same
        as the original variable to reduce performance overhead.
        It will be a silent error (a dimension too small error from `dae.build_pattern`)
        if a model uses `ExtAlgeb` to access a `State`, or vice versa.

        """

        self.parent = ext_model

        if isinstance(ext_model, GroupBase):
            if self.indexer.n > 0 and isinstance(self.indexer.v[0], (list, np.ndarray)):
                self._n = [len(i) for i in self.indexer.v]  # number of elements in each sublist
                self._idx = np.concatenate([np.array(i) for i in self.indexer.v])
            else:
                self._n = [len(self.indexer.v)]
                self._idx = self.indexer.v

            # use `0` for non-existent addresses (corr. to None in indexer)
            self.a = ext_model.get(src=self.src,
                                   idx=self._idx,
                                   attr='a',
                                   allow_none=self.allow_none,
                                   default=0,
                                   ).astype(int)
            self.n = len(self.a)

        else:
            original_var = ext_model.__dict__[self.src]

            if self.allow_none:
                raise NotImplementedError(f"{self.name}: allow_none not implemented for Model")

            if self.indexer is not None:
                uid = ext_model.idx2uid(self.indexer.v)
            else:
                uid = np.arange(ext_model.n, dtype=int)

            self._n = [len(uid)]
            if len(uid) > 0:
                self.a = original_var.a[uid]
            else:
                self.a = np.array([], dtype=int)

            # set initial v and e values to zero
            self.n = len(self.a)

        self.v = np.zeros(self.n)
        self.e = np.zeros(self.n)


class ExtState(ExtVar):
    """
    External state variable type.

    Warnings
    --------
    ``ExtState`` is not allowed to set ``t_const``, as it will conflict with the
    source ``State`` variable. In fact, one should not set ``e_str`` for ``ExtState``.
    """
    e_code = 'f'
    v_code = 'x'
    t_const = None


class ExtAlgeb(ExtVar):
    """
    External algebraic variable type.
    """
    e_code = 'g'
    v_code = 'y'



from typing import Optional, Union, Callable, Type
from andes.core.param import BaseParam
from andes.utils.func import list_flatten
from andes.core.common import dummify
from andes.shared import np, ndarray
import logging
from collections import OrderedDict
from andes.utils.tab import Tab

logger = logging.getLogger(__name__)


class BaseService(object):
    """
    Base class for Service.

    Service is a v-provider type for holding internal and temporary values. Subclasses need to implement ``v``
    as a member attribute or using a property decorator.

    Parameters
    ----------
    name : str
        Instance name

    Attributes
    ----------
    owner : Model
        The hosting/owner model instance
    """

    def __init__(self, name: str = None, tex_name: str = None, info: str = None, vtype: Type = None):
        self.name = name
        self.tex_name = tex_name if tex_name else name
        self.info = info
        self.vtype = vtype if vtype is not None else np.float  # type for `v`
        self.owner = None

    def get_names(self):
        """
        Return `name` in a list

        Returns
        -------
        list
            A list only containing the name of the service variable
        """
        return [self.name]

    @property
    def n(self):
        """
        Return the count of values in ``self.v``.

        Needs to be overloaded if ``v`` of subclasses is not a 1-dimensional array.

        Returns
        -------
        int
            The count of elements in this variable
        """
        if isinstance(self.v, (list, np.ndarray)):
            return len(self.v)
        else:
            return 1

    @property
    def class_name(self):
        """
        Return the class name
        """
        return self.__class__.__name__


class ConstService(BaseService):
    """
    A type of Service that stays constant once initialized.

    ConstService are usually constants calculated from parameters. They are only evaluated once in the
    initialization phase before variables are initialized. Therefore, uninitialized variables must not be
    used in `v_str``.

    Parameters
    ----------
    name : str
        Name of the ConstService
    v_str : str
        An equation string to calculate the variable value.
    v_numeric : Callable, optional
        A callable which returns the value of the ConstService

    Attributes
    ----------
    v : array-like or a scalar
        ConstService value
    """

    def __init__(self,
                 v_str: Optional[str] = None,
                 v_numeric: Optional[Callable] = None,
                 vtype: Optional[type] = None,
                 name: Optional[str] = None, tex_name=None, info=None):
        super().__init__(name=name, vtype=vtype, tex_name=tex_name, info=info)
        self.v_str = v_str
        self.v_numeric = v_numeric
        self.v: Union[float, int, ndarray] = np.array([0.])

    def assign_memory(self, n):
        """Assign memory for ``self.v`` and set the array to zero."""
        self.v = np.zeros(n, dtype=self.vtype)


class VarService(ConstService):
    """
    Variable service that gets updated in each step/loop as variables change.

    This class is useful when one has non-differentiable algebraic equations,
    which make use of `abs()`, `re` and `im`.
    Instead of creating `Algeb`, one can put the equation in `VarService`,
    which will be updated before solving algebraic equations.

    Examples
    --------
    In ESST3A model, the voltage and current sensors (vd + jvq), (Id + jIq)
    estimate the sensed VE using equation

    .. math ::

        VE = | K_{PC}*(v_d + 1j v_q) + 1j (K_I + K_{PC}*X_L)*(I_d + 1j I_q)|

    One can use `VarService` to implement this equation ::

        self.VE = VarService(tex_name='V_E',
                             info='VE',
                             v_str='Abs(KPC*(vd + 1j*vq) + 1j*(KI + KPC*XL)*(Id + 1j*Iq))',
                             )

    Warnings
    --------
    `VarService` is not solved with other algebraic equations, meaning that
    there is one step "delay" between the algebraic variables and `VarService`.
    Use an algebraic variable whenever possible.
    """

    pass


class PostInitService(ConstService):
    """
    Constant service that gets stored once after init.

    This service is useful when one need to store initialization
    values stored in variables.

    Examples
    --------
    In ESST3A model, the `vf` variable is initialized followed by other
    variables. One can store the initial `vf` into `vf0` so that equation
    ``vf - vf0 = 0`` will hold. ::

        self.vref0 = PostInitService(info='Initial reference voltage input',
                                     tex_name='V_{ref0}',
                                     v_str='vref',
                                     )

    Since all `ConstService` are evaluated before equation evaluation,
    without using PostInitService, one will need to create lots
    of `ConstService` to store values in the initialization path
    towards `vf0`, in order to correctly initialize `vf`.
    """

    pass


class BackRef(BaseService):
    """
    A special type of reference collector.

    `BackRef` is used for collecting device indices of other models referencing the parent model of the
    `BackRef`. The `v``field will be a list of lists, each containing the `idx` of other models
    referencing each device of the parent model.

    BackRef can be passed as indexer for params and vars, or shape for `NumReduce` and
    `NumRepeat`. See examples for illustration.

    Examples
    --------
    A Bus device has an `IdxParam` of `area`, storing the `idx` of area to which the bus device belongs.
    In ``Bus.__init__()``, one has ::

        self.area = IdxParam(model='Area')

    Suppose `Bus` has the following data

        ====   ====  ====
        idx    area  Vn
        ----   ----  ----
        1      1     110
        2      2     220
        3      1     345
        4      1     500
        ====   ====  ====

    The Area model wants to collect the indices of Bus devices which points to the corresponding Area device.
    In ``Area.__init__``, one defines ::

        self.Bus = BackRef()

    where the member attribute name `Bus` needs to match exactly model name that `Area` wants to collect
    `idx` for.
    Similarly, one can define ``self.ACTopology = BackRef()`` to collect devices in the `ACTopology` group
    that references Area.

    The collection of `idx` happens in :py:func:`andes.system.System._collect_ref_param`.
    It has to be noted that the specific `Area` entry must exist to collect model idx-dx referencing it.
    For example, if `Area` has the following data ::

        idx
        1

    Then, only Bus 1, 3, and 4 will be collected into `self.Bus.v`, namely, ``self.Bus.v == [ [1, 3, 4] ]``.

    If `Area` has data ::

        idx
        1
        2

    Then, `self.Bus.v` will end up with ``[ [1, 3, 4], [2] ]``.

    See Also
    --------
    andes.core.service.NumReduce : A more complete example using BackRef to build the COI model

    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.export = False
        self.v = list()


class ExtService(BaseService):
    """
    Service constants whose value is from an external model or group.

    Parameters
    ----------
    src : str
        Variable or parameter name in the source model or group
    model : str
        A model name or a group name
    indexer : IdxParam or BaseParam
        An "Indexer" instance whose ``v`` field contains the ``idx`` of devices in the model or group.

    Examples
    --------
    A synchronous generator needs to retrieve the ``p`` and ``q`` values from static generators
    for initialization. ``ExtService`` is used for this purpose.

    In a synchronous generator, one can define the following to retrieve ``StaticGen.p`` as ``p0``::

        class GENCLSModel(Model):
            def __init__(...):
                ...
                self.p0 = ExtService(src='p',
                                     model='StaticGen',
                                     indexer=self.gen,
                                     tex_name='P_0')

    """

    def __init__(self,
                 model: str,
                 src: str,
                 indexer: Union[BaseParam, BaseService],
                 attr: str = 'v',
                 allow_none: bool = False,
                 default=0,
                 name: str = None,
                 tex_name: str = None,
                 vtype=None,
                 info: str = None,
                 ):
        super().__init__(name=name, tex_name=tex_name, info=info, vtype=vtype)
        self.model = model
        self.src = src
        self.indexer = indexer
        self.attr = attr
        self.allow_none = allow_none
        self.default = default
        self.v = np.array([0.])

    def assign_memory(self, n):
        """Assign memory for ``self.v`` and set the array to zero."""
        self.v = np.zeros(n, dtype=self.vtype)

    def link_external(self, ext_model):
        """
        Method to be called by ``System`` for getting values from the external model or group.

        Parameters
        ----------
        ext_model
            An instance of a model or group provided by System
        """
        # set initial v values to zero
        self.v = np.zeros(self.n)
        if self.n == 0:
            return

        # the same `get` api for Group and Model
        self.v = ext_model.get(src=self.src, idx=self.indexer.v, attr=self.attr,
                               allow_none=self.allow_none,
                               default=self.default,
                               )


class DataSelect(BaseService):
    """
    Class for selecting values for optional DataParam or NumParam.

    This service is a v-provider that uses optional DataParam if available with a fallback.

    DataParam will be tested for `None`, and NumParam will be tested with `np.isnan()`.

    Notes
    -----
    An use case of DataSelect is remote bus. One can do ::

        self.buss = DataSelect(option=self.busr, fallback=self.bus)

    Then, pass ``self.buss`` instead of ``self.bus`` as indexer to retrieve voltages.

    Another use case is to allow an optional turbine rating. One can do ::

        self.Tn = NumParam(default=None)
        self.Sg = ExtParam(...)
        self.Sn = DataSelect(Tn, Sg)

    """

    def __init__(self,
                 optional,
                 fallback,
                 name: Optional[str] = None,
                 tex_name: Optional[str] = None,
                 info: Optional[str] = None,
                 ):
        super().__init__(name=name, tex_name=tex_name, info=info, )
        self.optional = optional
        self.fallback = fallback
        self._v = None

    @property
    def v(self):
        if self._v is None:
            self._v = [v1 if v1 is not None and not np.isnan(v1)
                       else v2
                       for v1, v2 in zip(self.optional.v, self.fallback.v)]

        return self._v


class DeviceFinder(BaseService):
    """
    Service for finding indices of optionally linked devices.

    If not provided, `DeviceFinder` will add devices at the beginning of `System.setup`.

    Examples
    --------
    IEEEST stabilizer takes an optional `busf` (IdxParam) for specifying the connected BusFreq,
    which is needed for mode 6. To avoid reimplementing `BusFreq` within IEEEST, one can do

    .. code-block :: python

        self.busfreq = DeviceFinder(self.busf, link=self.buss, idx_name='bus')

    where `self.busf` is the optional input, `self.buss` is the bus indices that `busf` should measure,
    and `idx_name` is the name of a BusFreq parameter through which the measured bus indices are specified.
    For each `None` values in `self.busf`, a `BusFreq` is created to measure the corresponding bus in `self.buss`.

    That is, ``BusFreq.[idx_name].v = [link]``. `DeviceFinder` will find / create `BusFreq` devices so that
    the returned list of `BusFreq` indices are connected to `self.buss`, respectively.
    """

    def __init__(self, u, link, idx_name, name=None, tex_name=None, info=None):
        super().__init__(name=name, tex_name=tex_name, info=info)

        self.u = u
        self.model = u.model
        self.idx_name = idx_name

        if self.model is None:
            raise ValueError(f'{u.owner.class_name}.{u.name} must contain "model".')

        self.link = link

    def find_or_add(self, system):
        mdl = system.models[self.model]
        found_idx = mdl.find_idx((self.idx_name,), (self.link.v,),
                                 allow_none=True, default=None)

        action = False
        for ii, idx in enumerate(found_idx):
            if idx is None:
                action = True
                new_idx = system.add(self.model, {self.idx_name: self.link.v[ii]})
                self.u.v[ii] = new_idx

                logger.info(f"{self.owner.class_name} <{self.owner.idx.v[ii]}> "
                            f"added {self.model} <{new_idx}> "
                            f"on {self.idx_name} <{self.link.v[ii]}>")
            else:
                action = True
                self.u.v[ii] = idx

        if action:
            mdl.list2array()
            mdl.refresh_inputs()

    @property
    def v(self):
        return self.u.v


class OperationService(BaseService):
    """
    Base class for a type of Service which performs specific operations

    This class cannot be used by itself.

    See Also
    --------
    NumReduce : Service for Reducing linearly stored 2-D services into 1-D

    NumRepeat : Service for repeating 1-D NumParam/ v-array following a sub-pattern

    IdxRepeat : Service for repeating 1-D IdxParam/ v-list following a sub-pattern
    """

    def __init__(self,
                 name=None,
                 tex_name=None,
                 info=None,
                 ):
        self._v = None
        super().__init__(name=name, tex_name=tex_name, info=info, )
        self.v_str = None

    @property
    def v(self):
        """
        Return values stored in `self._v`. May be overloaded by subclasses.
        """
        return self._v

    @v.setter
    def v(self, value):
        self._v = value


class NumReduce(OperationService):
    """
    A helper Service type which reduces a linearly stored 2-D ExtParam into 1-D Service.

    NumReduce works with ExtParam whose `v` field is a list of lists. A reduce function
    which takes an array-like and returns a scalar need to be supplied. NumReduce calls the reduce
    function on each of the lists and return all the scalars in an array.

    Parameters
    ----------
    u : ExtParam
        Input ExtParam whose ``v`` contains linearly stored 2-dimensional values
    ref : BackRef
        The BackRef whose 2-dimensional shapes are used for indexing
    fun : Callable
        The callable for converting a 1-D array-like to a scalar

    Examples
    --------
    Suppose one wants to calculate the mean value of the ``Vn`` in one Area. In the ``Area`` class, one defines ::

        class AreaModel(...):
            def __init__(...):
                ...
                # backward reference from `Bus`
                self.Bus = BackRef()

                # collect the Vn in an 1-D array
                self.Vn = ExtParam(model='Bus',
                    src='Vn',
                    indexer=self.Bus)

                self.Vn_mean = NumReduce(u=self.Vn,
                    fun=np.mean,
                    ref=self.Bus)

    Suppose we define two areas, 1 and 2, the Bus data looks like

        ===   =====  ====
        idx    area  Vn
        ---   -----  ----
        1      1     110
        2      2     220
        3      1     345
        4      1     500
        ===   =====  ====

    Then, `self.Bus.v` is a list of two lists ``[ [1, 3, 4], [2] ]``.
    `self.Vn.v` will be retrieved and linearly stored as ``[110, 345, 500, 220]``.
    Based on the shape from `self.Bus`, :py:func:`numpy.mean`
    will be called on ``[110, 345, 500]`` and ``[220]`` respectively.
    Thus, `self.Vn_mean.v` will become ``[318.33, 220]``.

    """

    def __init__(self,
                 u,
                 ref: BackRef,
                 fun: Callable,
                 name=None,
                 tex_name=None,
                 info=None,
                 cache=True,
                 ):
        super().__init__(name=name, tex_name=tex_name, info=info)
        self.u = u
        self.ref = ref
        self.fun = fun
        self.cache = cache

    @property
    def v(self):
        """
        Return the reduced values from the reduction function in an array

        Returns
        -------
        The array ``self._v`` storing the reduced values
        """
        if self._v is not None and self.cache is True:
            return self._v

        if self._v is None:
            self._v = np.zeros(len(self.ref.v))

        idx = 0
        for i, v in enumerate(self.ref.v):
            self._v[i] = self.fun(self.u.v[idx:idx + len(v)])
            idx += len(v)
        return self._v


class NumRepeat(OperationService):
    r"""
    A helper Service type which repeats a v-provider's value based on the shape from a BackRef

    Examples
    --------
    NumRepeat was originally designed for computing the inertia-weighted average rotor speed (center of
    inertia speed). COI speed is computed with

    .. math ::
        \omega_{COI} = \frac{ \sum{M_i * \omega_i} } {\sum{M_i}}

    The numerator can be calculated with a mix of BackRef, ExtParam and ExtState. The denominator needs to be
    calculated with NumReduce and Service Repeat. That is, use NumReduce to calculate the sum,
    and use NumRepeat to repeat the summed value for each device.

    In the COI class, one would have

    .. code-block :: python

        class COIModel(...):
            def __init__(...):
                ...
                self.SynGen = BackRef()
                self.SynGenIdx = RefFlatten(ref=self.SynGen)
                self.M = ExtParam(model='SynGen',
                                  src='M',
                                  indexer=self.SynGenIdx)

                self.wgen = ExtState(model='SynGen',
                                     src='omega',
                                     indexer=self.SynGenIdx)

                self.Mt = NumReduce(u=self.M,
                                         fun=np.sum,
                                         ref=self.SynGen)

                self.Mtr = NumRepeat(u=self.Mt,
                                       ref=self.SynGen)

                self.pidx = IdxRepeat(u=self.idx,ref=self.SynGen)

    Finally, one would define the center of inertia speed as

    .. code-block :: python

        self.wcoi = Algeb(v_str='1', e_str='-wcoi')

        self.wcoi_sub = ExtAlgeb(model='COI',
                                 src='wcoi',
                                 e_str='M * wgen / Mtr',
                                 v_str='M / Mtr',
                                 indexer=self.pidx,
                                 )

    It is very worth noting that the implementation uses a trick to separate the average weighted sum into `n`
    sub-equations, each calculating the :math:`(M_i * \omega_i) / (\sum{M_i})`. Since all the variables are
    preserved in the sub-equation, the derivatives can be calculated correctly.

    """

    def __init__(self,
                 u,
                 ref,
                 **kwargs):
        super().__init__(**kwargs)
        self.u = u
        self.ref = ref

    @property
    def v(self):
        """
        Return the values of the repeated values in a sequential 1-D array

        Returns
        -------
        The array, ``self._v`` storing the repeated values
        """
        if self._v is None:
            self._v = np.zeros(len(list_flatten(self.ref.v)))

            idx = 0
            for i, v in enumerate(self.ref.v):
                self._v[idx:idx + len(v)] = self.u.v[i]
                idx += len(v)
            return self._v
        else:
            return self._v


class IdxRepeat(OperationService):
    """
    Helper class to repeat IdxParam.

    This class has the same functionality as :py:class:`andes.core.service.NumRepeat`
    but only operates on IdxParam, DataParam or NumParam.
    """

    def __init__(self,
                 u,
                 ref,
                 **kwargs):
        super().__init__(**kwargs)
        self.u = u
        self.ref = ref

    @property
    def v(self):
        if self._v is None:
            self._v = [''] * len(list_flatten(self.ref.v))
            idx = 0
            for i, v in enumerate(self.ref.v):
                for jj in range(idx, idx + len(v)):
                    self._v[jj] = self.u.v[i]
                idx += len(v)
            return self._v
        else:
            return self._v


class RefFlatten(OperationService):
    """
    A service type for flattening :py:class:`andes.core.service.BackRef` into a 1-D list.

    Examples
    --------
    This class is used when one wants to pass `BackRef` values as indexer.

    :py:class:`andes.models.coi.COI` collects referencing
    :py:class:`andes.models.group.SynGen` with

    .. code-block :: python

        self.SynGen = BackRef(info='SynGen idx lists', export=False)

    After collecting BackRefs, `self.SynGen.v` will become a two-level list of indices,
    where the first level correspond to each COI and the second level correspond to generators
    of the COI.

    Convert `self.SynGen` into 1-d as `self.SynGenIdx`, which can be passed as indexer for
    retrieving other parameters and variables

    .. code-block :: python

        self.SynGenIdx = RefFlatten(ref=self.SynGen)

        self.M = ExtParam(model='SynGen', src='M',
                          indexer=self.SynGenIdx, export=False,
                          )
    """

    def __init__(self, ref, **kwargs):
        super().__init__(**kwargs)
        self.ref = ref

    @property
    def v(self):
        return list_flatten(self.ref.v)


class NumSelect(OperationService):
    """
    Class for selecting values for optional NumParam.

    Notes
    -----
    One use case is to allow an optional turbine rating. One can do ::

        self.Tn = NumParam(default=None)
        self.Sg = ExtParam(...)
        self.Sn = DataSelect(Tn, Sg)

    """

    def __init__(self,
                 optional,
                 fallback,
                 name: Optional[str] = None,
                 tex_name: Optional[str] = None,
                 info: Optional[str] = None,
                 ):
        super().__init__(name=name, tex_name=tex_name, info=info)
        self.optional = optional
        self.fallback = fallback
        self._v = None

    @property
    def v(self):
        if self._v is None:
            self._v = [v1 if not np.isnan(v1)
                       else v2
                       for v1, v2 in zip(self.optional.v, self.fallback.v)]

            self._v = np.array(self._v)

        return self._v


class InitChecker(OperationService):
    """
    Class for checking init values against known typical values.

    Instances will be stored in `Model.services_post` and
    `Model.services_icheck`, which will be checked in
    `Model.post_init_check()` after initialization.

    Parameters
    ----------
    u
        v-provider to be checked
    lower : float, BaseParam, BaseVar, BaseService
        lower bound
    upper : float, BaseParam, BaseVar, BaseService
        upper bound
    equal : float, BaseParam, BaseVar, BaseService
        values that the value from `v_str` should equal
    not_equal : float, BaseParam, BaseVar, BaseService
        values that should not equal
    enable : bool
        True to enable checking

    Examples
    --------
    Let's say generator excitation voltages are known to be in
    the range of 1.6 - 3.0 per unit. One can add the following
    instance to `GENBase` ::

        self._vfc = InitChecker(u=self.vf,
                                info='vf range',
                                lower=1.8,
                                upper=3.0,
                                )

    `lower` and `upper` can also take v-providers instead of
    float values.

    One can also pass float values from Config to make it
    adjustable as in our implementation of ``GENBase._vfc``.
    """

    def __init__(self, u, lower=None, upper=None, equal=None, not_equal=None,
                 enable=True, error_out=False, **kwargs):
        super().__init__(**kwargs)
        self.u = u
        self.lower = dummify(lower) if lower is not None else None
        self.upper = dummify(upper) if upper is not None else None
        self.equal = dummify(equal) if equal is not None else None
        self.not_equal = dummify(not_equal) if not_equal is not None else None
        self.enable = enable
        self.error_out = error_out

    def check(self):
        """
        Check the bounds and equality conditions.
        """
        if not self.enable:
            return

        def _not_all_close(a, b):
            return np.logical_not(np.isclose(a, b))

        if self._v is None:
            self._v = np.zeros_like(self.u.v)

        checks = [(self.lower, np.less_equal, "violation of the lower limit", "limit"),
                  (self.upper, np.greater_equal, "violation of the upper limit", "limit"),
                  (self.equal, _not_all_close, 'should be equal', "expected"),
                  (self.not_equal, np.equal, 'should not be equal', "not expected")
                  ]

        for check in checks:
            limit = check[0]
            func = check[1]
            text = check[2]
            text2 = check[3]
            if limit is None:
                continue

            self.v[:] = np.logical_or(self.v, func(self.u.v, limit.v))

            pos = np.argwhere(func(self.u.v, limit.v)).ravel()

            if len(pos) == 0:
                continue
            idx = [self.owner.idx.v[i] for i in pos]
            lim_v = limit.v * np.ones(self.n)

            title = f'{self.owner.class_name} {self.info} {text}.'

            err_dict = OrderedDict([('idx', idx),
                                    ('values', self.u.v[pos]),
                                    (f'{text2}', lim_v[pos]),
                                    ])
            data = list(map(list, zip(*err_dict.values())))

            tab = Tab(title=title, data=data, header=list(err_dict.keys()))
            if self.error_out:
                logger.error(tab.draw())
            else:
                logger.warning(tab.draw())

        self.v[:] = np.logical_not(self.v)


class FlagValue(BaseService):
    """
    Class for flagging values that equal to the given value.

    By default, values that equal to `value` will be flagged as `0`.
    Non-matching values will be flagged as `1`.

    Parameters
    ----------
    u
        Input parameter
    value
        Value to flag. Can be None, string, or a number.
    flag : 0 by default, only 0 or 1 is accepted.
        The flag for the matched ones

    Warnings
    --------
    `FlagNotNone` can only be applied to `BaseParam` with `cache=True`.
    Applying to `Service` will fail unless `cache` is False (at a performance cost).
    """

    def __init__(self, u, value, flag=0, name=None, tex_name=None, info=None, cache=True):
        BaseService.__init__(self, name=name, tex_name=tex_name, info=info)
        if flag != 0.0 and flag != 1.0:
            raise ValueError(f"flag must be 0 or 1. The given flag = {flag}.")

        self.u = u
        self.value = value
        self.flag = flag
        self.flag_neg = 1 - flag
        self.cache = cache

        self._v = None

    @property
    def v(self):
        new = False
        if self._v is None:
            self._v = np.zeros_like(self.u.v, dtype=float)
            new = True

        if not self.cache or new:
            # need to do it element-wise since `self.u.v` can be a list
            self._v[:] = np.array([self.flag if i == self.value else self.flag_neg
                                   for i in self.u.v])
        return self._v


class Replace(BaseService):
    """
    Replace parameters with new values if the function returns True
    """

    def __init__(self, old_val, flt, new_val, name=None, tex_name=None, info=None, cache=True):
        BaseService.__init__(self, name=name, tex_name=tex_name, info=info)
        self.cache = cache
        self.filter = flt  # function
        self.old_val = old_val
        self.new_val = dummify(new_val)
        self._v = None

    @property
    def v(self):
        new = False
        if self._v is None or not self.cache:
            self._v = np.zeros_like(self.old_val.v, dtype=float)
            new = True

        if not self.cache or new:
            new_v = self.new_val.v * np.ones_like(self.old_val.v)
            flt = self.filter(self.old_val.v)
            self._v[:] = new_v * flt + self.old_val.v * (1 - flt)

        return self._v


class ParamCalc(BaseService):
    """
    Parameter calculation service.

    Useful to create parameters calculated instantly from existing ones.
    """

    def __init__(self, param1, param2, func, name=None, tex_name=None, info=None,
                 cache=True):
        BaseService.__init__(self, name=name, tex_name=tex_name, info=info)
        self.param1 = param1
        self.param2 = param2
        self.func = func
        self.cache = cache
        self._v = None

    @property
    def v(self):
        new = False
        if self._v is None:
            new = True
            self._v = np.zeros_like(self.param1.v, dtype=float)

        if not self.cache or new:
            self._v[:] = self.func(self.param1.v,
                                   self.param2.v)

        return self._v


class RandomService(BaseService):
    """
    A service type for generating random numbers.

    Parameters
    ----------
    name : str
        Name
    func : Callable
        A callable for generating the random variable.

    Warnings
    --------
    The value will be randomized every time it is accessed. Do not use it if the value needs to be stable for
    each simulation step.
    """

    def __init__(self, func=np.random.rand, **kwargs):
        super(RandomService, self).__init__(**kwargs)
        self.func = func

    @property
    def v(self):
        """
        This class has `v` wrapped by a property decorator.

        Returns
        -------
        array-like
            Randomly generated service variables
        """
        return np.random.rand(self.n)



from andes.core.param import ExtParam, NumParam, IdxParam
from andes.core.model import Model, ModelData
from andes.core.var import ExtAlgeb, Algeb
from andes.core.service import NumReduce, NumRepeat, BackRef, DeviceFinder
from andes.core.discrete import Sampling
from andes.shared import np
from andes.utils.tab import Tab
from collections import OrderedDict


class AreaData(ModelData):
    def __init__(self):
        super().__init__()


class Area(AreaData, Model):
    def __init__(self, system, config):
        AreaData.__init__(self)
        Model.__init__(self, system, config)
        self.group = 'Collection'
        self.flags.pflow = True
        self.flags.tds = True

        self.Bus = BackRef()
        self.ACTopology = BackRef()

        # --------------------Experiment Zone--------------------
        self.Vn = ExtParam(model='Bus', src='Vn', indexer=self.ACTopology, export=False)
        self.Vn_sum = NumReduce(u=self.Vn, fun=np.sum, ref=self.Bus)
        self.Vn_sum_rep = NumRepeat(u=self.Vn_sum, ref=self.Bus)

        self.a = ExtAlgeb(model='ACTopology', src='a', indexer=self.ACTopology,
                          info='Bus voltage angle')
        self.v = ExtAlgeb(model='ACTopology', src='v', indexer=self.ACTopology,
                          info='Bus voltage magnitude')

        # self.time = Algeb(e_str='time - dae_t', v_setter=True)

    def bus_table(self):
        """
        Return a formatted table with area idx and bus idx correspondence

        Returns
        -------
        str
            Formatted table

        """
        if self.n:
            header = ['Area ID', 'Bus ID']
            rows = [(i, j) for i, j in zip(self.idx.v, self.Bus.v)]
            return Tab(header=header, data=rows).draw()
        else:
            return ''


class ACEData(ModelData):
    """
    Area Control Error data
    """

    def __init__(self):
        ModelData.__init__(self)
        self.bus = IdxParam(model='Bus', info="bus idx for freq. measurement", mandatory=True)
        self.bias = NumParam(default=1.0, info='bias parameter', tex_name=r'\beta',
                             unit='MW/0.1Hz', power=True)

        self.busf = IdxParam(info='Optional BusFreq idx', model='BusFreq',
                             default=None)


class ACE(ACEData, Model):
    """
    Area Control Error model.

    Note: area idx is automatically retrieved from `bus`.
    """

    def __init__(self, system, config):
        ACEData.__init__(self)
        Model.__init__(self, system, config)

        self.flags.tds = True
        self.group = 'Calculation'

        self.config.add(OrderedDict([('freq_model', 'BusFreq'),
                                     ('interval', 4.0),
                                     ('offset', 0.0),
                                     ]))
        self.config.add_extra('_help', {'freq_model': 'default freq. measurement model',
                                        'interval': 'sampling time interval',
                                        'offset': 'sampling time offset'})

        self.config.add_extra('_alt', {'freq_model': ('BusFreq',)})

        self.area = ExtParam(model='Bus', src='area', indexer=self.bus, export=False)

        self.busf.model = self.config.freq_model
        self.busfreq = DeviceFinder(self.busf, link=self.bus, idx_name='bus')

        self.f = ExtAlgeb(model='FreqMeasurement', src='f', indexer=self.busfreq,
                          export=False, info='Bus frequency',
                          )

        self.fs = Sampling(self.f,
                           interval=self.config.interval,
                           offset=self.config.offset,
                           tex_name='f_s',
                           info='Sampled freq.',
                           )

        self.ace = Algeb(info='area control error', unit='MW (p.u.)',
                         tex_name='ace',
                         e_str='10 * bias * (fs_v - 1) - ace',
                         )



from andes.core.solver import Solver
from andes.core import Config
from collections import OrderedDict


class BaseRoutine(object):
    """
    Base routine class.

    Provides references to system, config, and solver.
    """

    def __init__(self, system=None, config=None):
        self.system = system
        self.config = Config(self.class_name)

        if config is not None:
            self.config.load(config)

        self.config.add(OrderedDict((('sparselib', 'klu'),
                                     ('linsolve', 0),
                                     )))
        self.config.add_extra("_help",
                              sparselib="linear sparse solver name",
                              linsolve="solve symbolic factorization each step (enable when KLU segfaults)",
                              )
        self.config.add_extra("_alt", sparselib=("klu", "umfpack", "spsolve", "cupy"),
                              linsolve=(0, 1),
                              )

        self.solver = Solver(sparselib=self.config.sparselib)

    @property
    def class_name(self):
        return self.__class__.__name__

    def doc(self, max_width=78, export='plain'):
        return self.config.doc(max_width, export)

    def init(self):
        pass

    def run(self, **kwargs):
        raise NotImplementedError

    def summary(self):
        raise NotImplementedError

    def report(self):
        raise NotImplementedError



import logging
from andes.core.model import Model, ModelData  # NOQA
from andes.core.param import IdxParam, DataParam, NumParam  # NOQA
from andes.core.var import Algeb, State, ExtAlgeb  # NOQA
from andes.core.service import ConstService  # NOQA
import numpy as np
logger = logging.getLogger(__name__)


class LineData(ModelData):
    def __init__(self):
        super().__init__()

        self.bus1 = IdxParam(model='Bus', info="idx of from bus")
        self.bus2 = IdxParam(model='Bus', info="idx of to bus")

        self.Sn = NumParam(default=100.0, info="Power rating", non_zero=True, tex_name=r'S_n')
        self.fn = NumParam(default=60.0, info="rated frequency", tex_name=r'f')
        self.Vn1 = NumParam(default=110.0, info="AC voltage rating", non_zero=True, tex_name=r'V_{n1}')
        self.Vn2 = NumParam(default=110.0, info="rated voltage of bus2", non_zero=True, tex_name=r'V_{n2}')

        self.r = NumParam(default=1e-8, info="line resistance", tex_name='r', z=True)
        self.x = NumParam(default=1e-8, info="line reactance", tex_name='x', z=True)
        self.b = NumParam(default=0.0, info="shared shunt susceptance", y=True)
        self.g = NumParam(default=0.0, info="shared shunt conductance", y=True)
        self.b1 = NumParam(default=0.0, info="from-side susceptance", tex_name='b_1')
        self.g1 = NumParam(default=0.0, info="from-side conductance", tex_name='g_1')
        self.b2 = NumParam(default=0.0, info="to-side susceptance", tex_name='b_2')
        self.g2 = NumParam(default=0.0, info="to-side conductance", tex_name='g_2')

        self.trans = NumParam(default=0, info="transformer branch flag")
        self.tap = NumParam(default=1.0, info="transformer branch tap ratio", tex_name='t_{ap}')
        self.phi = NumParam(default=0.0, info="transformer branch phase shift in rad", tex_name=r'\phi')

        self.owner = IdxParam(model='Owner', info="owner code")

        self.xcoord = DataParam(info="x coordinates")
        self.ycoord = DataParam(info="y coordinates")


class Line(LineData, Model):
    def __init__(self, system=None, config=None):
        LineData.__init__(self)
        Model.__init__(self, system, config)
        self.group = 'ACLine'
        self.flags.pflow = True

        self.a1 = ExtAlgeb(model='Bus', src='a', indexer=self.bus1, tex_name='a_1',
                           info='phase angle of the from bus')
        self.a2 = ExtAlgeb(model='Bus', src='a', indexer=self.bus2, tex_name='a_2',
                           info='phase angle of the to bus')
        self.v1 = ExtAlgeb(model='Bus', src='v', indexer=self.bus1, tex_name='v_1',
                           info='voltage magnitude of the from bus')
        self.v2 = ExtAlgeb(model='Bus', src='v', indexer=self.bus2, tex_name='v_2',
                           info='voltage magnitude of the to bus')

        self.gh = ConstService(tex_name='g_h')
        self.bh = ConstService(tex_name='b_h')
        self.gk = ConstService(tex_name='g_k')
        self.bk = ConstService(tex_name='b_k')

        self.yh = ConstService(tex_name='y_h', vtype=np.complex)
        self.yk = ConstService(tex_name='y_k', vtype=np.complex)
        self.yhk = ConstService(tex_name='y_{hk}', vtype=np.complex)

        self.ghk = ConstService(tex_name='g_{hk}')
        self.bhk = ConstService(tex_name='b_{hk}')

        self.gh.v_str = 'g1 + 0.5 * g'
        self.bh.v_str = 'b1 + 0.5 * b'
        self.gk.v_str = 'g2 + 0.5 * g'
        self.bk.v_str = 'b2 + 0.5 * b'

        self.yh.v_str = 'u * (gh + 1j * bh)'
        self.yk.v_str = 'u * (gk + 1j * bk)'
        self.yhk.v_str = 'u/((r+1e-8) + 1j*(x+1e-8))'

        self.ghk.v_str = 're(yhk)'
        self.bhk.v_str = 'im(yhk)'

        self.a1.e_str = 'u * (v1 ** 2 * (gh + ghk) / tap ** 2  - \
                              v1 * v2 * (ghk * cos(a1 - a2 - phi) + \
                                         bhk * sin(a1 - a2 - phi)) / tap)'

        self.v1.e_str = 'u * (-v1 ** 2 * (bh + bhk) / tap ** 2 - \
                              v1 * v2 * (ghk * sin(a1 - a2 - phi) - \
                                         bhk * cos(a1 - a2 - phi)) / tap)'

        self.a2.e_str = 'u * (v2 ** 2 * (gh + ghk) - \
                              v1 * v2 * (ghk * cos(a1 - a2 - phi) - \
                                         bhk * sin(a1 - a2 - phi)) / tap)'

        self.v2.e_str = 'u * (-v2 ** 2 * (bh + bhk) + \
                              v1 * v2 * (ghk * sin(a1 - a2 - phi) + \
                                         bhk * cos(a1 - a2 - phi)) / tap)'



import pandas as pd
import andes

andes.config_logger(stream_level=30)


def tsat_to_df(file):
    """
    Convert a TSAT XLS export to DataFrame.
    """
    return excel_to_df(file, offset_row=6)


def psse_to_df(file):
    """
    Convert a PSS/E xlsx export to DataFrame.
    """
    return excel_to_df(file, offset_row=1, offset_col=0, )


def excel_to_df(file, offset_row=6, offset_col=0):
    """
    Convert a XLS export to DataFrame

    TSAT offset is 6, PSS/E offset is 1
    """
    wb = pd.read_excel(file)

    df = wb.iloc[offset_row:, offset_col:]
    df.transpose()
    df.columns = ['Time', *range(len(df.columns) - 1)]
    return df


def plot_comparison(system, variable, tsat_data, ylabel, a=None,
                    tsat_header=None, a_tsat=None, scale=1,
                    psse_data=None, psse_header=None, a_psse=None,
                    show=True, left=None, right=None,
                    legend=True, title=None):
    """
    Plot and compare ANDES, TSAT and PSS/E results.
    """
    plot = system.TDS.plotter.plot
    plot_data = system.TDS.plotter.plot_data
    a_tsat = list(a) if a_tsat is None else a_tsat
    a_psse = list(a) if a_psse is None else a_psse

    fig, ax = plot(variable,
                   a=a,
                   ycalc=lambda x: scale * x,
                   ylabel=ylabel,
                   show=False,
                   line_styles=['-', '-.'],
                   left=left,
                   right=right,
                   legend=legend,
                   title=title,
                   )

    fig, ax = plot_data(tsat_data['Time'].to_numpy(),
                        tsat_data[a_tsat].to_numpy(),
                        xheader=['Time [s]'],
                        yheader=[tsat_header[i] for i in a_tsat] if tsat_header is not None else None,
                        fig=fig,
                        ax=ax,
                        line_styles=['--', ':'],
                        left=left,
                        right=right,
                        legend=legend,
                        show=show,
                        title=title,
                        )
    if psse_data is not None:
        fig, ax = plot_data(
            psse_data['Time'].to_numpy(),
            psse_data[a_psse].to_numpy(),
            xheader=['Time [s]'],
            yheader=[psse_header[i] for i in a_psse] if psse_header is not None else None,
            fig=fig, ax=ax,
            line_styles=[':', '-.'],
            legend=legend,
            left=left,
            right=right,
            title=title,
        )
    return fig, ax


def run_cmp(raw, dyr, fault_line, t1=1, t2=1.1, tf=20, tstep=1/60, no_output=True,
            reconnect=True):
    """
    Run a case study with a line trip and reconnect event.
    """
    ss = andes.main.load(raw, addfile=dyr, routine='TDS',
                         tf=tf, setup=False, no_output=no_output)

    # configure simulation parameters
    ss.config.warn_limits = 0
    ss.config.warn_abnormal = 0

    ss.add('Toggler', {'model': 'Line', 'dev': fault_line, 't': t1})
    if reconnect:
        ss.add('Toggler', {'model': 'Line', 'dev': fault_line, 't': t2})

    ss.setup()
    ss.PFlow.run()
    ss.TDS.init()

    ss.TDS.config.tf = 20
    ss.TDS.config.tstep = tstep
    ss.TDS.run()

    return ss



"""
Experimental Models
"""
from andes.core.model import ModelData, Model
from andes.core.var import Algeb, State
from andes.core.param import NumParam
from andes.core.discrete import HardLimiter, AntiWindup  # NOQA
from andes.core.block import DeadBand1, PIController, PITrackAW, PIFreeze  # NOQA


class PI2Data(ModelData):
    """Data for PI2 model with deadlock issue"""
    def __init__(self):
        ModelData.__init__(self)
        self.Kp = NumParam()
        self.Ki = NumParam()
        self.Wmax = NumParam()
        self.Wmin = NumParam()


class PI2Model(Model):
    def __init__(self, system, config):
        Model.__init__(self, system, config)
        self.group = 'Experimental'
        self.flags.update({'tds': True})
        self.uin = State(v_str=0,
                         e_str='Piecewise((0, dae_t<= 0), (1, dae_t <= 2), (-1, dae_t <6), (1, True))',
                         tex_name='u_{in}',
                         )
        self.x = State(e_str='uin * Ki * HL_zi',
                       v_str=0.05,
                       )
        self.y = Algeb(e_str='uin * Kp + x - y',
                       v_str=0.05)

        self.HL = HardLimiter(u=self.y, lower=self.Wmin, upper=self.Wmax)
        self.w = Algeb(e_str='HL_zi * y + HL_zl * Wmin + HL_zu * Wmax - w',
                       v_str=0.05)


class PI2(PI2Data, PI2Model):
    def __init__(self, system, config):
        PI2Data.__init__(self)
        PI2Model.__init__(self, system, config)


class TestDB1(ModelData, Model):
    """
    Test model for `DeadBand1`.
    """
    def __init__(self, system, config):
        ModelData.__init__(self)
        Model.__init__(self, system, config)
        self.group = 'Experimental'

        self.flags.tds = True

        self.uin = Algeb(v_str='-10',
                         e_str='(dae_t - 10) - uin',
                         tex_name='u_{in}'
                         )
        self.DB = DeadBand1(self.uin, center=0, lower=-5, upper=5)


class TestPI(ModelData, Model):
    def __init__(self, system, config):
        ModelData.__init__(self)
        Model.__init__(self, system, config)
        self.group = 'Experimental'
        self.flags.tds = True

        self.uin = Algeb(v_str=0,
                         e_str='sin(dae_t) - uin',
                         tex_name='u_{in}',
                         )

        self.zf = Algeb(v_str=0,
                        e_str='Piecewise((0, dae_t <= 2), (1, dae_t <=6), (0, True)) - zf',
                        tex_name='z_f',
                        )

        self.PI = PIController(u=self.uin, kp=1, ki=0.1)

        self.PIAW = PITrackAW(u=self.uin, kp=0.5, ki=0.5, ks=2,
                              lower=-0.5, upper=0.5, x0=0.0,
                              )

        self.PIF = PIFreeze(u=self.uin, kp=0.5, ki=0.5, x0=0,
                            freeze=self.zf)

    def get_times(self):
        return (2.0, 6.0)



from andes.io.em_psse import parse_raw

import logging
logger = logging.getLogger(__name__)

raw2andes = {}


def is_format(fid):
    """
    Check the raw file for frequency base
    """
    first = fid.readline()
    first = first.strip().split('/')
    first = first[0].split(',')
    if float(first[5]) == 50.0 or float(first[5]) == 60.0:
        return True
    else:
        return False


def read(system, file_name):
    raw_data = parse_raw(file_name)
    for psse_model in raw_data:
        andes_model = raw2andes.get_model(psse_model)  # NOQA
        df = raw_data[psse_model]['df']
        for psse_param in df:
            pass

    return raw_data



"""
https://github.com/8080labs/pyforest/blob/master/src/pyforest/_importable.py

The MIT License (MIT)

Copyright (c) 2019 8080 Labs

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

"""
# Attention: LazyImport uses __double_dunder_methods__ instead of _var
# because we want to minimize the risk of collision.
# Collision happens if the imported module/object has an attribute with the same name.
# The result of collision is that the attribute of the imported module/object is masked.
# So, we are trading the risk of masking Python __builtins__ with masking module attributes
# because it will be easier to monitor the Python __builtins__
# than to monitor all attributes of libraries that might be imported with pyforest.
# Read more about dunder methods here: https://dbader.org/blog/meaning-of-underscores-in-python


class LazyImport(object):
    def __init__(self, import_statement):
        self.__import_statement__ = import_statement
        # the next line does not work for general imports, e.g. "from pandas import *"
        self.__imported_name__ = import_statement.strip().split()[-1]

        self.__complementary_imports__ = []
        self.__was_imported__ = False

    def __on_import__(self, lazy_import):
        self.__complementary_imports__.append(lazy_import)

    def __maybe_import_complementary_imports__(self):
        for lazy_import in self.__complementary_imports__:
            try:
                lazy_import.__maybe_import__()
            except:  # NOQA
                # silently fails if the complementary lazy_import is not available.
                # This is because complementary lazy_imports are considered optional.
                # Please note that direct imports of lazy_imports will fail explicitly.
                pass

    # Python will only import the module(s) if they are missing
    # if the module(s) were imported before, this method returns immediately
    def __maybe_import__(self):
        self.__maybe_import_complementary_imports__()
        exec(self.__import_statement__, globals())
        # Attention: if the import fails, the next line will not be reached
        self.__was_imported__ = True

    # among others, called during auto-completion of IPython/Jupyter
    def __dir__(self):
        self.__maybe_import__()
        return eval(f"dir({self.__imported_name__})")

    # called for undefined attribute and returns the attribute of the imported module
    def __getattr__(self, attribute):
        self.__maybe_import__()
        return eval(f"{self.__imported_name__}.{attribute}")

    # called for callable objects, e.g. from pathlib import Path; Path(".")
    def __call__(self, *args, **kwargs):
        self.__maybe_import__()
        return eval(self.__imported_name__)(*args, **kwargs)

    def __repr__(self, *args, **kwargs):
        # it is important that __repr__ does not trigger an import if the lazy_import is not yet imported
        # e.g. if the user calls locals(), this triggers __repr__ for each object
        # and this would result in an import if the lazy_import is not yet imported
        # and those imports might fail explicitly if the lazy_import is not available
        # and this would break locals() for the user

        if self.__was_imported__:
            # next line only adds imported_name into the local scope but does not trigger a new import
            # because the lazy_import was already imported via another trigger
            self.__maybe_import__()
            return f"active pyforest.LazyImport of {eval(self.__imported_name__)}"
        else:
            return f"lazy pyforest.LazyImport for '{self.__import_statement__}'"


def _import_statements(symbol_dict, was_imported=True):
    statements = []
    for _, symbol in symbol_dict.items():
        if isinstance(symbol, LazyImport) and (symbol.__was_imported__ == was_imported):
            print(symbol.__import_statement__)
            statements.append(symbol.__import_statement__)
    return statements



import argparse
import sys
import os
import logging
import platform
import importlib

from time import strftime
from andes.main import config_logger, find_log_path
from andes.utils.paths import get_log_dir
from andes.routines import routine_cli

logger = logging.getLogger(__name__)


def create_parser():
    """
    The main level of command-line interface.
    """
    parser = argparse.ArgumentParser()

    parser.add_argument(
        '-v', '--verbose',
        help='Program logging level in 10-DEBUG, 20-INFO, 30-WARNING, '
             '40-ERROR or 50-CRITICAL.',
        type=int, default=20, choices=(1, 10, 20, 30, 40, 50))

    sub_parsers = parser.add_subparsers(dest='command', help='[run] run simulation routine; '
                                                             '[plot] plot simulation results; '
                                                             '[doc] quick documentation; '
                                                             '[prepare] run the symbolic-to-numeric preparation; '
                                                             '[misc] miscellaneous functions.'
                                        )

    run = sub_parsers.add_parser('run')
    run.add_argument('filename', help='Case file name. Power flow is calculated by default.', nargs='*')
    run.add_argument('-r', '--routine', nargs='*', default=('pflow', ),
                     action='store', help='Simulation routine(s). Single routine or multiple separated with '
                                          'space. Run PFlow by default.',
                     choices=list(routine_cli.keys()))
    run.add_argument('-p', '--input-path', help='Path to case files', type=str, default='')
    run.add_argument('-a', '--addfile', help='Additional files used by some formats.')
    run.add_argument('-P', '--pert', help='Perturbation file path', default='')
    run.add_argument('-o', '--output-path', help='Output path prefix', type=str, default='')
    run.add_argument('-n', '--no-output', help='Force no output of any kind', action='store_true')
    run.add_argument('--ncpu', help='Number of parallel processes', type=int, default=os.cpu_count())
    run.add_argument('--dime', help='Specify DiME streaming server address and port', type=str)
    run.add_argument('--tf', help='End time of time-domain simulation', type=float)
    run.add_argument('-c', '--convert', help='Convert to format.', type=str, default='', nargs='?')
    run.add_argument('-b', '--add-book', help='Add a template workbook for the specified model.', type=str)
    run.add_argument('--convert-all', help='Convert to format with all templates.', type=str, default='',
                     nargs='?')
    run.add_argument('--state-matrix', help='Export state matrix to a .mat file. Need to run with `-r eig`',
                     action='store_true')
    run.add_argument('--profile', action='store_true', help='Enable Python cProfiler')
    run.add_argument('--shell', action='store_true', help='Start in IPython shell')
    run.add_argument('--no-preamble', action='store_true', help='Hide preamble')

    plot = sub_parsers.add_parser('plot')
    plot.add_argument('filename', nargs=1, default=[], help='simulation output file name, which should end '
                                                            'with `out`. File extension can be omitted.')
    plot.add_argument('x', nargs='?', type=int, help='the X-axis variable index, typically 0 for Time',
                      default='0')
    plot.add_argument('y', nargs='*', help='Y-axis variable indices. Space-separated indices or a '
                                           'colon-separated range is accepted')
    plot.add_argument('--xmin', type=float, help='minimum value for X axis', dest='left')
    plot.add_argument('--xmax', type=float, help='maximum value for X axis', dest='right')
    plot.add_argument('--ymax', type=float, help='maximum value for Y axis')
    plot.add_argument('--ymin', type=float, help='minimum value for Y axis')
    find_or_xargs = plot.add_mutually_exclusive_group()
    find_or_xargs.add_argument('--find', type=str, help='find variable indices that matches the given pattern')
    find_or_xargs.add_argument('-a', '--xargs', type=str,
                               help='find variable indices and return as a list of arguments '
                                    'usable with "|xargs andes plot"')
    plot.add_argument('--exclude', type=str, help='pattern to exclude in find or xargs results')
    plot.add_argument('-x', '--xlabel', type=str, help='x-axis label text')
    plot.add_argument('-y', '--ylabel', type=str, help='y-axis label text')
    plot.add_argument('-s', '--savefig', action='store_true', help='save figure. The default format is `png`')
    plot.add_argument('-f', '--format', dest='save_format',
                      help='format for savefig. Common formats such as png, pdf, jpg are supported')
    plot.add_argument('--dpi', type=int, help='image resolution in dot per inch (DPI)')
    plot.add_argument('--font-size', type=float, help='Text font size', default=12)
    plot.add_argument('--line-width', type=float, help='Plot line width', default=1.5)
    plot.add_argument('-g', '--grid', action='store_true', help='grid on')
    plot.add_argument('-t', '--title', help='title text')
    plot.add_argument('--greyscale', action='store_true', help='greyscale on')
    plot.add_argument('-d', '--no-latex', action='store_false', dest='latex', help='disable LaTex formatting')
    plot.add_argument('-n', '--no-show', action='store_false', dest='show', help='do not show the plot window')
    plot.add_argument('--ytimes', type=str, help='scale the y-axis values by YTIMES')
    plot.add_argument('-c', '--tocsv', help='convert npy output to csv', action='store_true')
    plot.add_argument('--hline1', help='dashed horizontal line 1', type=float)
    plot.add_argument('--hline2', help='dashed horizontal line 2', type=float)
    plot.add_argument('--vline1', help='dashed vertical line 1', type=float)
    plot.add_argument('--vline2', help='dashed vertical line 2', type=float)


    doc = sub_parsers.add_parser('doc')  # NOQA
    doc.add_argument('attribute', help='System attribute name to get documentation', nargs='?')
    doc.add_argument('--config', '-c', help='Config help')
    doc.add_argument('--list', '-l', help='List supported models and groups', action='store_true',
                     dest='list_supported')
    doc.add_argument('--init-seq', help='Show model initialization sequence', action='store_true',
                     )

    misc = sub_parsers.add_parser('misc')
    config_exclusive = misc.add_mutually_exclusive_group()
    config_exclusive.add_argument('--edit-config', help='Quick edit of the config file',
                                  default='', nargs='?', type=str)
    config_exclusive.add_argument('--save-config', help='save configuration to file name',
                                  nargs='?', type=str, default='')
    misc.add_argument('--license', action='store_true', help='Display software license', dest='show_license')
    misc.add_argument('-C', '--clean', help='Clean output files', action='store_true')
    misc.add_argument('-r', '--recursive', help='Recursively clean outputs (combined useage with --clean)',
                      action='store_true')

    prep = sub_parsers.add_parser('prepare')  # NOQA
    prep.add_argument('-q', '--quick', action='store_true', help='quick processing by skipping pretty prints')
    prep.add_argument('-i', '--incremental', action='store_true', help='Only incrementally generate for updated '
                                                                       'models')

    selftest = sub_parsers.add_parser('selftest')  # NOQA

    return parser


def preamble():
    """
    Log the ANDES command-line preamble at the `logging.INFO` level
    """
    from andes import __version__ as version

    py_version = platform.python_version()
    system_name = platform.system()
    date_time = strftime('%m/%d/%Y %I:%M:%S %p')
    logger.info("\n"
                rf"    _           _         | Version {version}" + '\n'
                rf"   /_\  _ _  __| |___ ___ | Python {py_version} on {system_name}, {date_time}" + '\n'
                r"  / _ \| ' \/ _` / -_|_-< | " + "\n"
                r' /_/ \_\_||_\__,_\___/__/ | This program comes with ABSOLUTELY NO WARRANTY.' + '\n')

    log_path = find_log_path(logging.getLogger("andes"))
    if len(log_path):
        logger.debug(f'Logging to file {log_path[0]}')


def main():
    """Main command-line interface"""
    parser = create_parser()
    args = parser.parse_args()

    config_logger(stream=True,
                  stream_level=args.verbose,
                  file=True,
                  log_path=get_log_dir(),
                  )
    logger.debug(args)

    module = importlib.import_module('andes.main')

    if args.command in ('plot', 'doc'):
        pass
    elif args.command == 'run' and args.no_preamble is True:
        pass
    else:
        preamble()

    if args.command is None:
        parser.parse_args(sys.argv.append('--help'))

    else:
        func = getattr(module, args.command)
        return func(cli=True, **vars(args))



import logging
from collections import OrderedDict
from andes.core.model import Model, ModelData
from andes.core.param import IdxParam, DataParam, NumParam
from andes.core.var import Algeb, ExtAlgeb, State  # NOQA
logger = logging.getLogger(__name__)


class NodeData(ModelData):
    def __init__(self):
        """
        DC Node data.
        """
        super().__init__()
        self.Vdcn = NumParam(default=100,
                             info='DC voltage rating',
                             unit='kV',
                             non_zero=True,
                             tex_name='V_{dcn}',
                             )
        self.Idcn = NumParam(default=1,
                             info='DC current rating',
                             unit='kA',
                             non_zero=True,
                             tex_name='I_{dcn}',
                             )
        self.v0 = NumParam(default=1.0,
                           info="initial voltage magnitude",
                           tex_name=r'V_{dc0}',
                           unit='p.u.',
                           )
        self.xcoord = DataParam(default=0,
                                info='x coordinate (longitude)',
                                )
        self.ycoord = DataParam(default=0,
                                info='y coordinate (latitude)',
                                )
        self.area = IdxParam(model='Area',
                             default=None,
                             info="Area code",
                             )
        self.zone = IdxParam(model='Region',
                             default=None,
                             info="Zone code",
                             )
        self.owner = IdxParam(model='Owner',
                              default=None,
                              info="Owner code",
                              )


class Node(NodeData, Model):
    """
    DC Node model.
    """
    def __init__(self, system, config):
        NodeData.__init__(self)
        Model.__init__(self, system=system, config=config)

        self.config.add(OrderedDict((('flat_start', 0),
                                     )))
        self.config.add_extra("_help",
                              flat_start="flat start for voltages",
                              )
        self.config.add_extra("_alt",
                              flat_start=(0, 1),
                              )
        self.config.add_extra("_tex",
                              flat_start="z_{flat}",
                              )

        self.group = 'DCTopology'
        self.category = ['TransNode']
        self.flags.update({'pflow': True})

        self.v = Algeb(name='v',
                       tex_name='V_{dc}',
                       info='voltage magnitude',
                       unit='p.u.',
                       diag_eps=1e-6,
                       )

        self.v.v_str = 'flat_start*1 + ' \
                       '(1-flat_start)*v0'


class DC2Term(ModelData, Model):
    """Two-terminal DC device template"""
    def __init__(self, system, config):
        ModelData.__init__(self)
        self.node1 = IdxParam(default=None,
                              info='Node 1 index',
                              mandatory=True,
                              model='Node',
                              )
        self.node2 = IdxParam(default=None,
                              info='Node 2 index',
                              mandatory=True,
                              model='Node',
                              )
        self.Vdcn1 = NumParam(default=100,
                              info='DC voltage rating on node 1',
                              unit='kV',
                              non_zero=True,
                              tex_name='V_{dcn1}',
                              )
        self.Vdcn2 = NumParam(default=100,
                              info='DC voltage rating on node 2',
                              unit='kV',
                              non_zero=True,
                              tex_name='V_{dcn2}',
                              )
        self.Idcn = NumParam(default=1,
                             info='DC current rating',
                             unit='kA',
                             non_zero=True,
                             tex_name='I_{dcn}',
                             )

        Model.__init__(self, system, config)
        self.v1 = ExtAlgeb(model='Node',
                           src='v',
                           indexer=self.node1,
                           info='DC voltage on node 1',
                           )
        self.v2 = ExtAlgeb(model='Node',
                           src='v',
                           indexer=self.node2,
                           info='DC voltage on node 2',
                           )


class ACDC2Term(ModelData, Model):
    """AC to two-terminal DC device template"""
    def __init__(self, system, config):
        ModelData.__init__(self)
        self.bus = IdxParam(model='Bus',
                            info="idx of connected bus",
                            mandatory=True,
                            )
        self.node1 = IdxParam(default=None,
                              info='Node 1 index',
                              mandatory=True,
                              model='Node',
                              )
        self.node2 = IdxParam(default=None,
                              info='Node 2 index',
                              mandatory=True,
                              model='Node',
                              )
        self.Vn = NumParam(default=110.0,
                           info="AC voltage rating",
                           non_zero=True,
                           tex_name=r'V_n',
                           )
        self.Vdcn1 = NumParam(default=100,
                              info='DC voltage rating on node 1',
                              unit='kV',
                              non_zero=True,
                              tex_name='V_{dcn1}',
                              )
        self.Vdcn2 = NumParam(default=100,
                              info='DC voltage rating on node 2',
                              unit='kV',
                              non_zero=True,
                              tex_name='V_{dcn2}',
                              )
        self.Idcn = NumParam(default=1,
                             info='DC current rating',
                             unit='kA',
                             non_zero=True,
                             tex_name='I_{dcn}',
                             )

        Model.__init__(self, system, config)
        self.a = ExtAlgeb(model='Bus',
                          src='a',
                          indexer=self.bus,
                          info='AC bus voltage phase',
                          )
        self.v = ExtAlgeb(model='Bus',
                          src='v',
                          indexer=self.bus,
                          info='AC bus voltage magnitude',
                          )
        self.v1 = ExtAlgeb(model='Node',
                           src='v',
                           indexer=self.node1,
                           info='DC node 1 voltage',
                           )
        self.v2 = ExtAlgeb(model='Node',
                           src='v',
                           indexer=self.node2,
                           info='DC node 2 voltage',
                           )


class Ground(ModelData, Model):
    """
    Ground model that sets the voltage of the connected DC node.
    """

    def __init__(self, system, config):
        ModelData.__init__(self)
        self.node = IdxParam(default=None,
                             info='Node index',
                             mandatory=True,
                             model='Node',
                             )
        self.voltage = NumParam(default=0.0,
                                tex_name='V_0',
                                info='Ground voltage (typically 0)',
                                unit='p.u.',
                                )
        Model.__init__(self, system, config)
        self.flags.update({'pflow': True})
        self.group = 'DCLink'
        self.v = ExtAlgeb(model='Node',
                          src='v',
                          indexer=self.node,
                          e_str='-Idc',
                          )
        self.Idc = Algeb(tex_name='I_{dc}',
                         info='Ficticious current injection from ground',
                         e_str='u * (v - voltage)',
                         v_str='0',
                         diag_eps=1e-6,
                         )
        self.v.e_str = '-Idc'


class R(DC2Term):
    """
    Resistive dc line
    """
    def __init__(self, system, config):
        DC2Term.__init__(self, system, config)
        self.flags.pflow = True
        self.group = 'DCLink'
        self.R = NumParam(unit='p.u.',
                          info='DC line resistance',
                          non_zero=True,
                          default=0.01,
                          r=True,
                          )
        self.Idc = Algeb(tex_name='I_{dc}',
                         info='Current from node 2 to 1',
                         unit='p.u.',
                         v_str='u * (v2 - v1) / R',
                         e_str='u * (v2 - v1) / R - Idc',
                         )
        self.v1.e_str = '-Idc'
        self.v2.e_str = '+Idc'


class L(DC2Term):
    """
    Inductive dc line
    """
    def __init__(self, system, config):
        DC2Term.__init__(self, system, config)
        self.flags.pflow = True
        self.group = 'DCLink'

        self.L = NumParam(unit='p.u.',
                          info='DC line inductance',
                          non_zero=True,
                          default=0.001,
                          r=True,
                          )
        self.IL = State(tex_name='I_L',
                        info='Inductance current',
                        unit='p.u.',
                        v_str='0',
                        e_str='-u * (v1 - v2) / L',
                        )
        self.v1.e_str = '-IL'
        self.v2.e_str = '+IL'


class C(DC2Term):
    """
    Capacitive dc branch
    """
    def __init__(self, system, config):
        DC2Term.__init__(self, system, config)
        self.flags.pflow = True
        self.group = 'DCLink'

        self.C = NumParam(unit='p.u.',
                          info='DC capacitance',
                          non_zero=True,
                          default=0.001,
                          g=True,
                          )
        self.vC = State(tex_name='v_C',
                        info='Capacitor current',
                        unit='p.u.',
                        v_str='0',
                        e_str='-u * Idc / C',
                        )
        self.Idc = Algeb(tex_name='I_{dc}',
                         info='Current from node 2 to 1',
                         unit='p.u.',
                         v_str='0',
                         e_str='u * (vC - (v1 - v2)) + '
                               '(1 - u) * Idc',
                         diag_eps=1e-6,
                         )
        self.v1.e_str = '-Idc'
        self.v2.e_str = '+Idc'


class RLs(DC2Term):
    def __init__(self, system, config):
        DC2Term.__init__(self, system, config)
        self.flags.pflow = True
        self.group = 'DCLink'

        self.R = NumParam(unit='p.u.',
                          tex_name='R',
                          info='DC line resistance',
                          non_zero=True,
                          default=0.01,
                          r=True,
                          )
        self.L = NumParam(unit='p.u.',
                          tex_name='L',
                          info='DC line inductance',
                          non_zero=True,
                          default=0.001,
                          r=True,
                          )
        self.IL = State(tex_name='I_L',
                        info='Inductance current',
                        unit='p.u.',
                        e_str='u * (v1 - v2 - R * IL) / L',
                        v_str='(v1 - v2) / R',
                        )
        self.Idc = Algeb(tex_name='I_{dc}',
                         info='Current from node 2 to 1',
                         unit='p.u.',
                         e_str='-u * IL - Idc',
                         v_str='-u * (v1 - v2) / R',
                         )
        self.v1.e_str = '-Idc'
        self.v2.e_str = '+Idc'


class RCp(DC2Term):
    def __init__(self, system, config):
        DC2Term.__init__(self, system, config)
        self.flags.pflow = True
        self.group = 'DCLink'

        self.R = NumParam(unit='p.u.',
                          tex_name='R',
                          info='DC line resistance',
                          non_zero=True,
                          default=0.01,
                          r=True,
                          )
        self.C = NumParam(unit='p.u.',
                          tex_name='C',
                          info='DC capacitance',
                          non_zero=True,
                          default=0.001,
                          g=True,
                          )
        self.vC = State(tex_name='v_C',
                        info='Capacitor current',
                        unit='p.u.',
                        e_str='-u * (Idc - vC/R) / C',
                        v_str='v1 - v2',
                        )
        self.Idc = Algeb(tex_name='I_{dc}',
                         info='Current from node 2 to 1',
                         unit='p.u.',
                         e_str='u * (vC - (v1 - v2)) + '
                               '(1 - u) * Idc',
                         v_str='-(v1 - v2) / R',
                         diag_eps=1e-6,
                         )
        self.v1.e_str = '-Idc'
        self.v2.e_str = '+Idc'


class RLCp(DC2Term):
    def __init__(self, system, config):
        DC2Term.__init__(self, system, config)
        self.flags.pflow = True
        self.group = 'DCLink'

        self.R = NumParam(unit='p.u.',
                          tex_name='R',
                          info='DC line resistance',
                          non_zero=True,
                          default=0.01,
                          r=True,
                          )
        self.L = NumParam(unit='p.u.',
                          tex_name='L',
                          info='DC line inductance',
                          non_zero=True,
                          default=0.001,
                          r=True,
                          )
        self.C = NumParam(unit='p.u.',
                          tex_name='C',
                          info='DC capacitance',
                          non_zero=True,
                          default=0.001,
                          g=True,
                          )
        self.IL = State(tex_name='I_L',
                        info='Inductance current',
                        unit='p.u.',
                        v_str='0',
                        e_str='u * vC / L',
                        )
        self.vC = State(tex_name='v_C',
                        info='Capacitor current',
                        unit='p.u.',
                        e_str='-u * (Idc - vC/R - IL) / C',
                        v_str='v1 - v2',
                        )
        self.Idc = Algeb(tex_name='I_{dc}',
                         info='Current from node 2 to 1',
                         unit='p.u.',
                         e_str='u * (vC - (v1 - v2)) + '
                               '(1 - u) * Idc',
                         v_str='-(v1 - v2) / R',
                         diag_eps=1e-6,
                         )
        self.v1.e_str = '-Idc'
        self.v2.e_str = '+Idc'


class RCs(DC2Term):
    def __init__(self, system, config):
        DC2Term.__init__(self, system, config)
        self.flags.pflow = True
        self.group = 'DCLink'

        self.R = NumParam(unit='p.u.',
                          tex_name='R',
                          info='DC line resistance',
                          non_zero=True,
                          default=0.01,
                          r=True,
                          )
        self.C = NumParam(unit='p.u.',
                          tex_name='C',
                          info='DC capacitance',
                          non_zero=True,
                          default=0.001,
                          g=True,
                          )
        self.vC = State(tex_name='v_C',
                        info='Capacitor current',
                        unit='p.u.',
                        e_str='-u * Idc / C',
                        v_str='v1 - v2',
                        )
        self.Idc = Algeb(tex_name='I_{dc}',
                         info='Current from node 2 to 1',
                         unit='p.u.',
                         e_str='u * (vC - (v1 - v2) - Idc * R) + '
                               '(1 - u) * Idc',
                         v_str='-(v1 - v2) / R',
                         diag_eps=1e-6,
                         )
        self.v1.e_str = '-Idc'
        self.v2.e_str = '+Idc'


class RLCs(DC2Term):
    def __init__(self, system, config):
        DC2Term.__init__(self, system, config)
        self.flags.pflow = True
        self.group = 'DCLink'

        self.R = NumParam(unit='p.u.',
                          tex_name='R',
                          info='DC line resistance',
                          non_zero=True,
                          default=0.01,
                          r=True,
                          )
        self.L = NumParam(unit='p.u.',
                          tex_name='L',
                          info='DC line inductance',
                          non_zero=True,
                          default=0.001,
                          r=True,
                          )
        self.C = NumParam(unit='p.u.',
                          tex_name='C',
                          info='DC capacitance',
                          non_zero=True,
                          default=0.001,
                          g=True,
                          )
        self.IL = State(tex_name='I_L',
                        info='Inductance current',
                        unit='p.u.',
                        e_str='u * (v1 - v2 - R * IL - vC) / L',
                        v_str='0',
                        )
        self.vC = State(tex_name='v_C',
                        info='Capacitor current',
                        unit='p.u.',
                        e_str='u * IL / C',
                        v_str='v1 - v2',
                        )
        self.Idc = Algeb(tex_name='I_{dc}',
                         info='Current from node 2 to 1',
                         unit='p.u.',
                         e_str='-IL - Idc',
                         v_str='0',
                         diag_eps=1e-6,
                         )
        self.v1.e_str = '-Idc'
        self.v2.e_str = '+Idc'



import unittest
import os
import andes
from andes.utils.paths import get_case

andes.main.config_logger(stream_level=30, file=True)


class TestMATPOWER(unittest.TestCase):

    def setUp(self) -> None:
        self.cases = ('case5.m', 'case14.m', 'case300.m', 'case118.m')

    def test_pflow_mpc_pool(self):
        case_path = [get_case(os.path.join('matpower', item)) for item in self.cases]
        andes.run(case_path, no_output=True, ncpu=2, pool=True, verbose=40)

    def test_pflow_mpc_process(self):
        case_path = [get_case(os.path.join('matpower', item)) for item in self.cases]
        andes.run(case_path, no_output=True, ncpu=2, pool=False, verbose=40)



from andes.core.param import TimerParam, IdxParam, DataParam, NumParam
from andes.core.model import Model, ModelData
from andes.core.var import ExtAlgeb
from andes.core.service import ConstService
from andes.shared import np, tqdm
import logging
logger = logging.getLogger(__name__)


class TogglerData(ModelData):
    def __init__(self):
        super(TogglerData, self).__init__()
        self.model = DataParam(info='Model or Group of the device to control', mandatory=True)
        self.dev = IdxParam(info='idx of the device to control', mandatory=True)
        self.t = TimerParam(info='switch time for connection status', mandatory=True)


class Toggler(TogglerData, Model):
    """
    Time-based connectivity status toggler.
    """
    def __init__(self, system, config):
        TogglerData.__init__(self)
        Model.__init__(self, system, config)
        self.flags.update({'tds': True})
        self.group = 'TimedEvent'

        self.t.callback = self._u_switch

    def _u_switch(self, is_time: np.ndarray):
        action = False
        for i in range(self.n):
            if is_time[i] and (self.u.v[i] == 1):
                instance = self.system.__dict__[self.model.v[i]]
                u0 = instance.get(src='u', attr='v', idx=self.dev.v[i])
                instance.set(src='u', attr='v', idx=self.dev.v[i], value=1-u0)
                action = True
                tqdm.write(f'<Toggler {self.idx.v[i]}>: '
                           f'{self.model.v[i]}.{self.dev.v[i]} status changed to {1-u0} at t={self.t.v[i]} sec.')
        return action


class Fault(ModelData, Model):
    """
    Three-phase to ground fault.
    """
    def __init__(self, system, config):
        ModelData.__init__(self)
        self.bus = IdxParam(model='Bus',
                            info="linked bus idx",
                            mandatory=True,
                            )
        self.tf = TimerParam(info='Bus fault start time',
                             unit='second',
                             mandatory=True,
                             callback=self.apply_fault,
                             )
        self.tc = TimerParam(info='Bus fault end time',
                             unit='second',
                             callback=self.clear_fault,
                             )
        self.xf = NumParam(info='Fault to ground impedance (positive)',
                           unit='p.u.(sys)',
                           default=1e-4,
                           tex_name='x_f',
                           )
        self.rf = NumParam(info='Fault to ground resistance (positive)',
                           unit='p.u.(sys)',
                           default=0,
                           tex_name='x_f',
                           )

        Model.__init__(self, system, config)
        self.flags.update({'tds': True})
        self.group = 'TimedEvent'
        self.gf = ConstService(tex_name='g_{f}',
                               v_str='re(1/(rf + 1j * xf))',
                               vtype=np.complex,
                               )
        self.bf = ConstService(tex_name='b_{f}',
                               v_str='im(1/(rf + 1j * xf))',
                               vtype=np.complex,
                               )

        # uf: an internal flag of whether the fault is in action (1) or not (0)
        self.uf = ConstService(tex_name='u_f', v_str='0')

        self.a = ExtAlgeb(model='Bus',
                          src='a',
                          indexer=self.bus,
                          tex_name=r'\theta',
                          info='Bus voltage angle',
                          unit='p.u.(kV)',
                          e_str='u * uf * (v ** 2 * gf)',
                          )
        self.v = ExtAlgeb(model='Bus',
                          src='v',
                          indexer=self.bus,
                          tex_name=r'V',
                          unit='p.u.(kV)',
                          info='Bus voltage magnitude',
                          e_str='-u * uf * (v ** 2 * bf)',
                          )
        self._vstore = np.array([])

    def apply_fault(self, is_time: np.ndarray):
        """
        Apply fault and store pre-fault algebraic variables (voltages and other algebs) to `self._vstore`.
        """
        action = False
        for i in range(self.n):
            if is_time[i] and (self.u.v[i] == 1):
                self.uf.v[i] = 1
                self._vstore = np.array(self.system.dae.y[self.system.Bus.n:])
                tqdm.write(f'<Fault {self.idx.v[i]}>: '
                           f'Applying fault on Bus (idx={self.bus.v[i]}) at t={self.tf.v[i]}sec.')
                action = True
        return action

    def clear_fault(self, is_time: np.ndarray):
        """
        Clear fault and restore pre-fault bus algebraic variables (voltages and others).
        """
        action = False
        for i in range(self.n):
            if is_time[i] and (self.u.v[i] == 1):
                self.uf.v[i] = 0
                self.system.dae.y[self.system.Bus.n:] = self._vstore
                tqdm.write(f'<Fault {self.idx.v[i]}>: '
                           f'Clearing fault on Bus (idx={self.bus.v[i]}) at t={self.tc.v[i]}sec.')
                action = True
        return action



from andes.core.model import Model, ModelData
from andes.core.param import NumParam, IdxParam, ExtParam
from andes.core.block import Piecewise, Lag, GainLimiter, LagAntiWindupRate
from andes.core.var import ExtAlgeb, Algeb
from andes.core.service import ConstService, FlagValue, ExtService


class REGCAU1Data(ModelData):
    """
    REGC_A model data.
    """
    def __init__(self):
        ModelData.__init__(self)

        self.bus = IdxParam(model='Bus',
                            info="interface bus id",
                            mandatory=True,
                            )
        self.gen = IdxParam(info="static generator index",
                            mandatory=True,
                            )
        self.Tg = NumParam(default=0.1, tex_name='T_g',
                           info='converter time const.', unit='s',
                           )
        self.Rrpwr = NumParam(default=999, tex_name='R_{rpwr}',
                              info='Low voltage power logic (LVPL) ramp limit',
                              unit='p.u.',
                              )
        self.Brkpt = NumParam(default=1.0, tex_name='B_{rkpt}',
                              info='LVPL characteristic voltage 2',
                              unit='p.u.',
                              )
        self.Zerox = NumParam(default=0.5, tex_name='Z_{erox}',
                              info='LVPL characteristic voltage 1',
                              unit='p.u',
                              )
        # TODO: ensure Brkpt > Zerox
        self.Lvpl1 = NumParam(default=1.0, tex_name='L_{vpl1}',
                              info='LVPL gain',
                              unit='p.u',
                              )
        self.Volim = NumParam(default=1.2, tex_name='V_{olim}',
                              info='Voltage lim for high volt. reactive current mgnt.',
                              unit='p.u.',
                              )
        self.Lvpnt1 = NumParam(default=1.0, tex_name='L_{vpnt1}',
                               info='High volt. point for low volt. active current mgnt.',
                               unit='p.u.',
                               )
        self.Lvpnt0 = NumParam(default=0.4, tex_name='L_{vpnt0}',
                               info='Low volt. point for low volt. active current mgnt.',
                               unit='p.u.',
                               )
        # TODO: ensure Lvpnt1 > Lvpnt0
        self.Iolim = NumParam(default=999.0, tex_name='I_{olim}',
                              info='current limit for high volt. reactive current mgnt.',
                              unit='p.u.',
                              )
        self.Tfltr = NumParam(default=0.1, tex_name='T_{fltr}',
                              info='Voltage filter T const for low volt. active current mgnt.',
                              unit='s',
                              )
        self.Khv = NumParam(default=0.7, tex_name='K_{hv}',
                            info='Overvolt. compensation gain in high volt. reactive current mgnt.',
                            )
        self.Iqrmax = NumParam(default=999, tex_name='I_{qrmax}',
                               info='Upper limit on the ROC for reactive current',
                               unit='p.u.',
                               )
        self.Iqrmin = NumParam(default=-999, tex_name='I_{qrmin}',
                               info='Lower limit on the ROC for reactive current',
                               unit='p.u.',
                               )
        self.Accel = NumParam(default=0.0, tex_name='A_{ccel}',
                              info='Acceleration factor',
                              vrange=(0, 1.0),
                              )
        self.Iqmax = NumParam(default=999, tex_name='I_{qmax}',
                              info='Upper limit for reactive current',
                              unit='p.u.',
                              )
        self.Iqmin = NumParam(default=-999, tex_name='I_{qmin}',
                              info='Lower limit for reactive current',
                              unit='p.u.',
                              )


class REGCAU1Model(Model):
    """
    REGCAU1 implementation.
    """
    def __init__(self, system, config):
        Model.__init__(self, system, config)

        self.a = ExtAlgeb(model='Bus',
                          src='a',
                          indexer=self.bus,
                          tex_name=r'\theta',
                          info='Bus voltage angle',
                          e_str='Ipout * v',
                          )

        self.v = ExtAlgeb(model='Bus',
                          src='v',
                          indexer=self.bus,
                          tex_name=r'V',
                          info='Bus voltage magnitude',
                          e_str='Iqout_y * v',
                          )

        self.p0 = ExtService(model='StaticGen',
                             src='p',
                             indexer=self.gen,
                             tex_name='P_0',
                             )
        self.q0 = ExtService(model='StaticGen',
                             src='q',
                             indexer=self.gen,
                             tex_name='Q_0',
                             )
        self.ra = ExtParam(model='StaticGen',
                           src='ra',
                           indexer=self.gen,
                           tex_name='r_a',
                           )
        self.xs = ExtParam(model='StaticGen',
                           src='xs',
                           indexer=self.gen,
                           tex_name='x_s',
                           )

        # --- INITIALIZATION ---
        self.Ipcmd0 = ConstService('p0 / v', info='initial Ipcmd')

        self.Iqcmd0 = ConstService('-q0 / v', info='initial Iqcmd')

        self.Ipcmd = Algeb(tex_name='I_{pcmd}',
                           info='current component for active power',
                           e_str='Ipcmd0 - Ipcmd', v_str='Ipcmd0')

        self.Iqcmd = Algeb(tex_name='I_{qcmd}',
                           info='current component for reactive power',
                           e_str='Iqcmd0 - Iqcmd', v_str='Iqcmd0')

        # reactive power management

        # TODO: create conditions for rate limiting.
        #   In a fault recovery, activate upper limit when Qg0 > 0
        #                        activate lower limit when Qg0 < 0

        self.S1 = LagAntiWindupRate(u=self.Iqcmd, T=self.Tg, K=-1,
                                    lower=self.Iqmin, upper=self.Iqmax,
                                    rate_lower=self.Iqrmin, rate_upper=self.Iqrmax,
                                    # rate_lower_cond, rate_upper_cond,
                                    tex_name='S_1',
                                    info='Iqcmd delay',
                                    )  # output `S1_y` == `Iq`

        # piece-wise gain for low voltage reactive current mgnt.
        self.kLVG = ConstService(v_str='1 / (Lvpnt1 - Lvpnt0)',
                                 tex_name='k_{LVG}',
                                 )

        self.LVG = Piecewise(u=self.v, points=('Lvpnt0', 'Lvpnt1'),
                             funs=('0', '(v - Lvpnt0) * kLVG', '1'),
                             info='Low voltage current gain',
                             )

        self.Lvplsw = FlagValue(u=self.Lvpl1, value=0, flag=0, tex_name='z_{Lvplsw}',
                                info='LVPL enable switch',
                                )
        # piece-wise gain for LVPL
        self.kLVPL = ConstService(v_str='Lvplsw * Lvpl1 / (Brkpt - Zerox)',
                                  tex_name='k_{LVPL}',
                                  )

        self.S2 = Lag(u=self.v, T=self.Tfltr, K=1.0,
                      info='Voltage filter with no anti-windup',
                      )
        self.LVPL = Piecewise(u=self.S2_y,
                              points=('Zerox', 'Brkpt'),
                              funs=('0 + 9999*(1-Lvplsw)',
                                    '(S2_y - Zerox) * kLVPL + 9999 * (1-Lvplsw)',
                                    '9999'),
                              info='Low voltage Ipcmd upper limit',
                              )

        self.S0 = LagAntiWindupRate(u=self.Ipcmd, T=self.Tg, K=1,
                                    upper=self.LVPL_y, rate_upper=self.Rrpwr,
                                    lower=-999, rate_lower=-999,
                                    no_lower=True, rate_no_lower=True,
                                    )  # `S0_y` is the output `Ip` in the block diagram

        self.Ipout = Algeb(e_str='S0_y * LVG_y -Ipout',
                           v_str='Ipcmd * LVG_y',
                           info='Output Ip current',
                           )

        # high voltage part
        self.HVG = GainLimiter(u='v - Volim', K=self.Khv, info='High voltage gain block',
                               lower=0, upper=999, no_upper=True)

        self.Iqout = GainLimiter(u='S1_y- HVG_y', K=1, lower=self.Iolim, upper=9999,
                                 no_upper=True, info='Iq output block')  # `Iqout_y` is the final Iq output

    def v_numeric(self, **kwargs):
        """
        Disable the corresponding `StaticGen`s.
        """
        self.system.groups['StaticGen'].set(src='u', idx=self.gen.v, attr='v', value=0)


class REGCAU1(REGCAU1Data, REGCAU1Model):
    def __init__(self, system, config):
        REGCAU1Data.__init__(self)
        REGCAU1Model.__init__(self, system, config)


class REECA1Data(ModelData):
    """
    Renewable energy electrical control model REEC_A data.
    """
    pass


class REECA1Model(Model):
    """
    REEC_A model implementation.

    Completed:
      1. Dead band type 1, implement and test (implemented and tested (TestDB1))

    TODO:
      2. PI controller with state freeze (implemented)
      2.1 PI controller with anti-windup limiter and state freeze (implemented)
      2.2 v_drop signal generator that creates a switching event (adds time to `t_switch`).
      3. Lag with state freeze (implemented)
      3.1 Lag with anti-windup limiter with state freeze (implemented)
      4. Nonlinear blocks `VDL1` and `VDL2`
      5. Value and time-based state transition

    """
    pass



from .jit import JIT  # NOQA
from collections import OrderedDict  # NOQA


non_jit = OrderedDict([
    ('bus', ['Bus']),
    ('pq', ['PQ']),
    ('pv', ['PV', 'Slack']),
    ('shunt', ['Shunt']),
    ('line', ['Line']),
    ('area', ['Area', 'ACE']),
    ('synchronous', ['GENCLS', 'GENROU']),
    ('governor', ['TG2', 'TGOV1', 'IEEEG1']),
    ('exciter', ['EXDC2', 'IEEEX1', 'ESDC2A', 'EXST1', 'ESST3A', 'SEXS']),
    ('pss', ['IEEEST', 'ST2CUT']),
    ('measurement', ['BusFreq', 'BusROCOF', 'PMU']),
    ('coi', ['COI', ]),
    ('dcbase', ['Node', 'Ground', 'R', 'L', 'C', 'RCp', 'RCs', 'RLs', 'RLCs', 'RLCp']),
    ('vsc', ['VSCShunt']),
    ('wtg', ['REGCAU1']),
    ('timer', ['Toggler', 'Fault']),
    ('experimental', ['PI2', 'TestDB1', 'TestPI']),
])



import logging
from _pydecimal import Decimal, ROUND_DOWN
from time import time

logger = logging.getLogger(__name__)


def elapsed(t0=0.0):
    """
    Get the elapsed time from the give time.
    If the start time is not given, returns the unix-time.

    Returns
    -------
    t : float
        Elapsed time from the given time; Otherwise the epoch time.
    s : str
        The elapsed time in seconds in a string
    """
    t = time()
    dt = t - t0
    dt_sec = Decimal(str(dt)).quantize(Decimal('.0001'), rounding=ROUND_DOWN)
    if dt_sec == 1:
        s = str(dt_sec) + ' second'
    else:
        s = str(dt_sec) + ' seconds'
    return t, s


def to_number(s):
    """
    Convert a string to a number. If unsuccessful, return the de-blanked string.
    """
    ret = s

    # remove single quotes
    if "'" in ret:
        ret = ret.strip("'").strip()

    # try converting to booleans / None
    if ret == 'True':
        return True
    elif ret == 'False':
        return False
    elif ret == 'None':
        return None

    # try converting to float or int
    try:
        ret = int(ret)
    except ValueError:
        try:
            ret = float(ret)
        except ValueError:
            pass

    return ret


def is_notebook():
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True   # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False      # Probably standard Python interpreter


def is_interactive():
    import __main__ as main
    return not hasattr(main, '__file__')



"""
Shared constants and delayed imports.

PAST NOTES
**********
Known issues of LazyImport ::

    1) The delayed import of pandas and newton_krylov will cause a ``RuntimeWarning``

    RuntimeWarning: numpy.ufunc size changed, may indicate binary incompatibility. Expected 192 from C header,
    got 216 from PyObject
        return f(*args, **kwds)

    2) High overhead when called hundreds of thousands times. For example, NumPy must not be imported with
    LazyImport.

    3) Prevents from serialization due to recursion depth.
"""

import pandas as pd  # NOQA
import matplotlib as mpl  # NOQA
from matplotlib import pyplot as plt  # NOQA
from pathos.multiprocessing import Pool  # NOQA
from multiprocessing import Process  # NOQA
import unittest  # NOQA
import yaml  # NOQA

from scipy.optimize import newton_krylov  # NOQA
from scipy.optimize import fsolve  # NOQA
from scipy.integrate import solve_ivp  # NOQA
from scipy.integrate import odeint  # NOQA

import math
import coloredlogs  # NOQA
import numpy as np  # NOQA
from tqdm import tqdm  # NOQA

import cvxopt  # NOQA
from cvxopt import umfpack  # NOQA
from cvxopt import spmatrix, matrix, sparse, spdiag  # NOQA
from numpy import ndarray  # NOQA

from andes.utils.texttable import Texttable  # NOQA

try:
    from cvxoptklu import klu
except ImportError:
    klu = None


deg2rad = math.pi/180
jac_names = ('fx', 'fy', 'gx', 'gy')
jac_types = ('c', '')

jac_full_names = list()
for jname in jac_names:
    for jtype in jac_types:
        jac_full_names.append(jname + jtype)

# *** Deprecated  ***
#
# # Packages
# pd = LazyImport('import pandas')
# plt = LazyImport('from matplotlib import pyplot')
# mpl = LazyImport('import matplotlib')
# Pool = LazyImport('from pathos.multiprocessing import Pool')
# unittest = LazyImport('import unittest')
# yaml = LazyImport('import yaml')
#
# # function calls
# newton_krylov = LazyImport('from scipy.optimize import newton_krylov')
# fsolve = LazyImport('from scipy.optimize import fsolve')
# solve_ivp = LazyImport('from scipy.integrate import solve_ivp')
# odeint = LazyImport('from scipy.integrate import odeint')
#
# *** Deprecated  ***
